"""

ECLiPt SSH Shell main module

"""

from sys import path, exit

path.append("/usr/lib/esshsh")

from tools import TRUE, FALSE

try:
    import gnome.applet
    appletsupport = TRUE
except:
    appletsupport = FALSE
    pass

from gtk import GtkButton, GtkVBox, WINDOW_POPUP, GtkWindow, GtkHandleBox, GtkHBox, timeout_remove, mainquit
from gtk import GtkScrolledWindow, POLICY_AUTOMATIC, GtkCList, DEST_DEFAULT_ALL, GdkColor, mainloop
from gtk import create_pixmap_from_xpm, GtkCTree, GtkLabel, JUSTIFY_CENTER, GtkPixmap, timeout_add
from gtk import GtkFrame, SHADOW_NONE

from gnome.ui import GnomeErrorDialog, STOCK_MENU_STOP, STOCK_MENU_PREF, STOCK_MENU_ABOUT, GnomeOkDialog
from gnome.ui import GnomeAbout, GnomeQuestionDialog, YES

from ConfigParser import ConfigParser
from re import sub
from string import strip, atoi, join, lower, find
from socket import gethostbyname, gethostname
from time import sleep

try:
    from twofish import twofish
except:
    print "Error - couldn't find twofish encryption module"
    exit()

from binascii import b2a_base64, a2b_base64
from os import unlink, environ
from tools import SigChldHandler, split, mkdirtree, cfill, bool2str, str2bool, merge

from constants import LISTVIEW, TREEVIEW, SSH, RSH, TELNET, VNC, IDEA, DES, TRIPLEDES, BLOWFISH, ARCFOUR
from constants import NONE, MODE_EXECUTE, hostdirectory, MODE_COPY, gladefile, MW_NONE, personaldirectory
from constants import appletversion, version, title, MW_SHORT, MW_LONG, CF_SHOWLIST
from constants import CF_PREFERENCES, targets, pixmappath, ressourcefile, defaultcommand
from constants import defaultforward, defaultrsh, defaultscp, defaulttelnet, defaultvncviewer
from constants import defaultwindowx, defaultwindowy, defaultwidthx, defaultwidthy, defaultappletsizex
from constants import defaultappletsizey, defaultappletpixmap, defaultbgone, defaultbgtwo
from constants import defaultcolorroot, defaultnonappletposx, defaultnonappletposy, defaultsshwinmode
from constants import defaultappletpixmapmouseover

import os
import signal
try:
    import expect
except:
    print "Error - couldn't find expect module"
    exit()

import sys
import libglade
import GDK
import log4py

# -----------------------------------------------------------------------------
# Miscellaneous functions

def viewmode2str(viewmode):
    if viewmode == LISTVIEW: return "ListView"
    elif viewmode == TREEVIEW: return "TreeView"

def str2viewmode(str):
    if lower(str) == "listview": return LISTVIEW
    elif lower(str) == "treeview": return TREEVIEW

def protocol2str(protocol):
    if protocol == SSH: return "SSH"
    elif protocol == RSH: return "RSH"
    elif protocol == TELNET: return "Telnet"
    elif protocol == VNC: return "VNC"

def str2protocol(str):
    if lower(str) == "ssh": return SSH
    elif lower(str) == "rsh": return RSH
    elif lower(str) == "telnet": return TELNET
    elif lower(str) == "vnc": return VNC

def cipher2str(cipher):
    if cipher == IDEA: return "IDEA"
    elif cipher == DES: return "DES"
    elif cipher == TRIPLEDES: return "3DES"
    elif cipher == BLOWFISH: return "Blowfish"
    elif cipher == ARCFOUR: return "ArcFour"
    elif cipher == NONE: return "None"
    
def str2cipher(str):
    if lower(str) == "idea": return IDEA
    elif lower(str) == "des": return DES
    elif lower(str) == "3des": return TRIPLEDES
    elif lower(str) == "blowfish": return BLOWFISH
    elif lower(str) == "arcfour": return ARCFOUR
    elif lower(str) == "none": return NONE

def itemname(name, category):
    return "%s-%s" % (category, name)

# -----------------------------------------------------------------------------
# Host object

class host:

    # Initialize host
    def __init__(self, defaultcipher = IDEA):

        # Save the host state itself
        self.name = ""
        self.category = ""
        self.hostname = ""
        self.username = ""
        self.password = ""
        self.internalpassword = FALSE
        self.protocol = SSH
        self.launch = ""
        self.sshport = 22
        self.cipher = defaultcipher
        self.x11forwarding = TRUE
        self.compression = FALSE
        self.localforwarding = ""
        self.remoteforwarding = ""
        self.defaultdirectory = ""
        self.telnetport = 23
        self.encodings = ""
        self.fullscreen = FALSE
        self.viewonly = FALSE
        self.shared = FALSE
        self.tunnel = FALSE
        self.zlib = ""
        self.application = ""

        # Internal variables
        self.filename = ""
        self.command = ""
        self.mode = MODE_EXECUTE
        self.log4py = log4py.Category(self)

    # Load one host file
    def Load(self, filename):
        self.filename = filename

        parser = ConfigParser()                             # The Configuration File Parser

        try:
            parser.read(self.filename)

            for i in range(len(parser.options("host"))):
                option = parser.options("host")[i]
                value = parser.get("host", option)

                # Load settings for the "Command" tab
                if (option == "connectionname"): self.name = value
                elif (option == "category"): self.category = value
                elif (option == "protocol"): self.protocol = str2protocol(value)
                elif (option == "hostname"): self.hostname = value
                elif (option == "username"): self.username = value
                elif (option == "internalpassword"): self.internalpassword = str2bool(value)
                elif (option == "password"): self.password = value
                elif (option == "application"): self.application = value
                elif (option == "defaultdirectory"): self.defaultdirectory = value

                # Save SSH host settings
                elif (option == "launch"): self.launch = value
                elif (option == "localforwarding"): self.localforwarding = value
                elif (option == "remoteforwarding"): self.remoteforwarding = value
                elif (option == "sshport"): self.sshport = atoi(value)
                elif (option == "cipher"): self.cipher = str2cipher(value)
                elif (option == "x11forwarding") : self.x11forwarding = str2bool(value)
                elif (option == "compression"): self.compression = str2bool(value)
        
                # Save Telnet host settings
                elif (option == "telnetport"): self.telnetport = atoi(value)

                # Savet VNC host settings
                elif (option == "encodings"): self.encodings = value
                elif (option == "viewonly"): self.viewonly = str2bool(value)
                elif (option == "fullscreen"): self.fullscreen = str2bool(value)
                elif (option == "shared"): self.shared = str2bool(value)
                elif (option == "tunnel"): self.tunnel = str2bool(value)
                elif (option == "zlib"): self.zlib = value

        except:
            self.log4py.error("Error while parsing host file.")

    # Save this host description file
    def Save(self):

        filename = "%s/%s-%s" % (hostdirectory, self.category, self.name)

        if (self.filename != filename):
            if os.path.exists(self.filename):
                unlink(self.filename)
        self.filename = filename

        file = open(self.filename, "w")
        file.write("[host]\n")
        
        # Save general host settings
        file.write("ConnectionName=%s\n" % self.name)
        file.write("Category=%s\n" % self.category)
        file.write("Protocol=%s\n" % protocol2str(self.protocol))
        file.write("Hostname=%s\n" % self.hostname)
        file.write("Username=%s\n" % self.username)
        file.write("InternalPassword=%s\n" % bool2str(self.internalpassword))
        file.write("Password=%s\n" % strip(self.password))
        file.write("Application=%s\n" % self.application)

        # Save SSH host settings
        file.write("Launch=%s\n" % self.launch)
        file.write("LocalForwarding=%s\n" % self.localforwarding)
        file.write("RemoteForwarding=%s\n" % self.remoteforwarding)
        file.write("SSHPort=%d\n" % self.sshport)
        file.write("Cipher=%s\n" % cipher2str(self.cipher))
        file.write("X11Forwarding=%s\n" % bool2str(self.x11forwarding))
        file.write("Compression=%s\n" % bool2str(self.compression))
        file.write("DefaultDirectory=%s\n" % self.defaultdirectory)

        # Save Telnet host settings
        file.write("TelnetPort=%d\n" % self.telnetport)

        # Savet VNC host settings
        file.write("Encodings=%s\n" % self.encodings)
        file.write("ZLib=%s\n" % self.zlib)
        file.write("ViewOnly=%s\n" % bool2str(self.viewonly))
        file.write("Fullscreen=%s\n" % bool2str(self.fullscreen))
        file.write("Shared=%s\n" % bool2str(self.shared))
        file.write("Tunnel=%s\n" % bool2str(self.tunnel))

        file.close()

    # Try to resolve the hostname
    def ResolveHostname(self):
        try:
            ip = gethostbyname(self.hostname)
            return TRUE
        except:
            GnomeErrorDialog("Couldn't resolve hostname \"%s\" !" % self.hostname)
            return FALSE

    # Substitute command line parameters
    def SubstituteParameters(self):

        self.command = sub("%c", cipher2str(self.cipher), self.command)
        if (self.compression == TRUE):
            self.command = sub("%C", "-C", self.command)
        else:
            self.command = sub("%C", "", self.command)
        self.command = sub("%f", self.launch, self.command)
        self.command = sub("%h", self.hostname, self.command)
        if (self.username != ""):
            self.command = sub("%u", "%s" % self.username, self.command)
        else:
            self.command = sub("%u", "", self.command)
            self.command = sub("-l", "", self.command)
            
        if (self.x11forwarding == TRUE):
            self.command = sub("%x", "-X", self.command)
        else:
            self.command = sub("%x", "-x", self.command)
        if (self.localforwarding != ""):
            splitted = split(self.localforwarding, " ")
            localcommand = ""
            for i in range(len(splitted)):
                localcommand = "%s -L %s" % (localcommand, splitted[i])
            self.command = sub("%L", localcommand, self.command)
        else:
            self.command = sub("%L", "", self.command)
        if (self.remoteforwarding != ""):
            splitted = split(self.remoteforwarding, " ")
            remotecommand = ""
            for i in range(len(splitted)):
                remotecommand = "%s -R %s" % (remotecommand, splitted[i])
            self.command = sub("%R", remotecommand, self.command)
        else:
            self.command = sub("%R", "", self.command)

        if (self.encodings != ""):
            self.command = sub("%e", "\"%s\"" % self.encodings, self.command)
        else:
            self.command = sub("%e", "", self.command)
            self.command = sub("-encodings", "", self.command)

        # VNC command settings

        if (self.fullscreen == TRUE):
            self.command = sub("%s", "-fullscreen", self.command)
        else:
            self.command = sub("%s", "", self.command)

        if (self.viewonly == TRUE):
            self.command = sub("%v", "-viewonly", self.command)
        else:
            self.command = sub("%v", "", self.command)

        if (self.shared == TRUE):
            self.command = sub("%S", "-shared", self.command)
        else:
            self.command = sub("%S", "", self.command)

        if (self.tunnel == TRUE):
            self.command = sub("%t", "-tunnel", self.command)
            environ["VNC_TUNNEL_CMD"] = "/usr/bin/ssh -f -L %%L:%%H:%%R %s@%%H sleep 20" % self.username
        else:
            self.command = sub("%t", "", self.command)

        if (self.zlib != ""):
            self.command = sub("%z", self.zlib, self.command)
        else:
            self.command = sub("%z", "", self.command)
            self.command = sub("-zlib", "", self.command)

        # Miscellaneous

        if (self.application != ""):
            self.command = sub("%a", self.application, self.command)
        else:
            self.command = sub("%a", "", self.command)

        self.command = strip(self.command)

    # Execute command specified by the host
    def Execute(self, sshcommand, sshforwarding, rshcommand, telnetcommand, vncviewercommand, resolvehostname, password = ""):

        # Try to resolve the hostname
        if (self.protocol != VNC):
            if (resolvehostname == TRUE):
                result = self.ResolveHostname()
                if (result == FALSE):
                    return FALSE

        # Set the appropriate command
        if (self.protocol == SSH):
            if (self.launch != ""):
                self.command = sshforwarding
            else:
                self.command = sshcommand
            self.command = sub("%p", str(self.sshport), self.command)
        elif (self.protocol == RSH):
            self.command = rshcommand
        elif (self.protocol == TELNET):
            self.command = telnetcommand
            self.command = sub("%p", str(self.telnetport), self.command)
        elif (self.protocol == VNC):
            self.command = vncviewercommand

        self.mode = MODE_EXECUTE
        self.SubstituteParameters()

        if (self.protocol == VNC) and (self.internalpassword == TRUE) and (password != ""):
            self.ForkExcpect(password)
        else:
            self.ForkExecv()

    # Fork and execute the command
    def ForkExecv(self):
        self.log4py.debug("Fork [ Execv ]: %s" % self.command)
        result = os.fork()
        if (result == 0):
            os.execv("/usr/bin/env", ["/usr/bin/env"] + split(self.command, " ", "\""))

    # Fork and use Expect to execute the command
    def ForkExcpect(self, password = ""):
        self.log4py.debug("Fork [ Expect ]: %s" % self.command)

        # [preisl@daddy ~/Software/E-SSH-Shell]$ scp ecat.py preisl@linux:
        # ecat.py                   |          0 KB |   0.5 kB/s | ETA: 00:00:00 | 100%

        result = os.fork()
        if (result == 0):
            ex = expect.popen2(self.command, "p")
            try:
                if (password != ""):
                    line = ex.read("word: ")
                    sleep(0.1)
                    ex.write("%s\n" % password)
                while 1:
                    if (ex.pid() == -1):
                        break
                    else:
                        line = ex.readline()
                        if (find(lower(line), "permission denied") != -1):
                            self.log4py.warn("Fork [ Expect ]: Permission denied (wrong password).")
                            os.kill(ex.pid(), 15)
                        else:
                            sys.stdout.write(line)

            except (IOError), error:
                self.log4py.error("%s" % error)
            os._exit(0)

    # Copy files by usinge the command specified by the host
    def CopyFile(self, scpcommand, resolvehostname, directory, files, password = ""):
        
        if (self.protocol == RSH) or (self.protocol == TELNET) or (self.protocol == VNC):
            GnomeErrorDialog("Copying files is only allowed for SSH connections !")
            return FALSE

        if (resolvehostname == TRUE):
            result = self.ResolveHostname()
            if (result == FALSE):
                return FALSE

        self.command = scpcommand
        self.command = sub("%p", str(self.sshport), self.command)

        self.mode = MODE_COPY
        self.SubstituteParameters()

        self.command = sub("%D", directory, self.command)
        self.command = sub("%F", files, self.command)

        if (self.internalpassword == TRUE):
            self.command = split(self.command, " ", "\"")[-1]
            self.ForkExcpect(password)
        else:
            self.ForkExecv()
        
# -----------------------------------------------------------------------------
# Main Object

class esshsh:

    # Initialize program
    def __init__(self, target):

        self.glade = libglade.GladeXML(gladefile)
        self.target = target
        self.sshwindowmode = 0
        self.currenthost = None
        self.locked = FALSE
        self.callingfunction = None
        self.mainwindowstatus = MW_NONE
        self.log4py = log4py.Category(self)

        self.hosts = {}                                            # List of hosts

        self.InitPreferences()
        self.InitHostPreferences()
        self.InitCopyFiles()

        mkdirtree(personaldirectory)
        mkdirtree(hostdirectory)

        self.InitCrypto()
        self.InitCreateMainWindow()

    # Initialize preferences window
    def InitPreferences(self):

        self.prefwin = self.glade.get_widget("preferences")
        self.prefwin.connect("destroy", self.PreferencesWindowHide)
        self.prefwin.connect("delete_event", self.PreferencesWindowHide)

        # Items on the "Command" tab
        self.sshcommand = self.glade.get_widget("sshcommand")
        self.sshforwarding = self.glade.get_widget("sshforwarding")
        self.rshcommand = self.glade.get_widget("rshcommand")
        self.telnetcommand = self.glade.get_widget("telnetcommand")
        self.scpcommand = self.glade.get_widget("scpcommand")
        self.vncviewercommand = self.glade.get_widget("vncviewercommand")
        self.helplist = self.glade.get_widget("helplist")

        # Items on the "Hosts" tab
        self.prefhostlist = self.glade.get_widget("prefhostlist")
        self.prefhostlist.set_auto_sort(TRUE)
        addButton = self.glade.get_widget("addhost")
        addButton.connect("clicked", self.AddHost)
        updateButton = self.glade.get_widget("updatehost")
        updateButton.connect("clicked", self.UpdateHost)
        removeButton = self.glade.get_widget("removehost")
        removeButton.connect("clicked", self.RemoveHost)

        # Items on the "Window Size & Position" tab
        self.windowsizex = self.glade.get_widget("windowsizex")
        self.windowsizey = self.glade.get_widget("windowsizey")
        self.windowpositionx = self.glade.get_widget("windowpositionx")
        self.windowpositiony = self.glade.get_widget("windowpositiony")
        self.appletsizex = self.glade.get_widget("appletsizex")
        self.appletsizey = self.glade.get_widget("appletsizey")
        self.appletpixmap = self.glade.get_widget("appletpixmap")
        self.appletpixmapmouseover = self.glade.get_widget("appletpixmapmouseover")
        self.bgcolorone = self.glade.get_widget("bgcolorone")
        self.bgcolortwo = self.glade.get_widget("bgcolortwo")
        self.rootcolor = self.glade.get_widget("rootcolor")
        self.treeview = self.glade.get_widget("treeview")
        self.listview = self.glade.get_widget("listview")
        self.nonappletpositionx = self.glade.get_widget("nonappletpositionx")
        self.nonappletpositiony = self.glade.get_widget("nonappletpositiony")

        # Items in the "Miscellaneous" tab
        self.resolve = self.glade.get_widget("resolve")
        self.defaultcipher = self.glade.get_widget("defaultcipher")

    # Initialize host preferences window
    def InitHostPreferences(self):

        self.hostwin = self.glade.get_widget("hostpreferences")
        self.hostwin.connect("destroy", self.HostPreferencesWindowHide)
        self.hostwin.connect("delete_event", self.HostPreferencesWindowHide)

        # Items in the "General" tab
        self.connectionname = self.glade.get_widget("connectionname")
        self.category = self.glade.get_widget("category")
        self.protocol = self.glade.get_widget("protocol")
        self.hostname = self.glade.get_widget("hostname")
        self.password = self.glade.get_widget("password")
        self.internalpassword = self.glade.get_widget("internalpassword")
        self.username = self.glade.get_widget("username")
        self.startapplication = self.glade.get_widget("startapplication")

        # Items in the "SSH" tab
        self.launch = self.glade.get_widget("launch")
        self.localforwarding = self.glade.get_widget("localforwarding")
        self.remoteforwarding = self.glade.get_widget("remoteforwarding")
        self.sshport = self.glade.get_widget("sshport")
        self.sshcipher = self.glade.get_widget("sshcipher")
        self.x11forwarding = self.glade.get_widget("x11forwarding")
        self.compression = self.glade.get_widget("compression")
        self.scpdirectory = self.glade.get_widget("scpdirectory")

        # Items in the "Telnet" tab
        self.telnetport = self.glade.get_widget("telnetport")

        # Items in the "VNC" tab
        self.encodings = self.glade.get_widget("encodings")
        self.zlib = self.glade.get_widget("zlib")
        self.viewonly = self.glade.get_widget("viewonly")
        self.fullscreen = self.glade.get_widget("fullscreen")
        self.shared = self.glade.get_widget("shared")
        self.tunnel = self.glade.get_widget("tunnel")

    # Initialize host preferences window
    def InitCopyFiles(self):
        self.copyfile = self.glade.get_widget("copyfiles")
        self.copyfile.connect("destroy", self.HideCopyFileWindow)
        self.copyfile.connect("delete_event", self.HideCopyFileWindow)
        startbutton = self.glade.get_widget("startBtn")
        startbutton.connect("clicked", self.CopyFile)
        self.directory = self.glade.get_widget("directory")
        self.copyfilelist = self.glade.get_widget("copyfilelist")
        self.hostinformation = self.glade.get_widget("hostinformation")

    # Initialize program after success full authentication
    def InitCreateMainWindow(self):
        self.LoadPreferences()
        signal.signal(signal.SIGCHLD, SigChldHandler)

        if (self.target == appletversion):
            if (appletsupport == FALSE):
                self.log4py.error("esshsh-applet: no pygnome-applet bindings installed.")
                exit(1)
            self.CreateApplet()
        else:
            self.CreateNonApplet(None)

        self.CreateHelpCList()                                # Insert help into the helplist

    # Highlight the applet icon when moving the mouse over it
    def HighlightApplet(self, *args):
        self.app_pix.set(self.pixbright, self.maskbright)

    # Darken the applet icon when mouse leaves the icon
    def DarkenApplet(self, *args):
        self.app_pix.set(self.pixdark, self.maskdark)

    # Create the applet
    def CreateApplet(self):
        self.applet = gnome.applet.AppletWidget("esshsh")
        self.applet.set_tooltip("%s %s" % (title, version))
        self.applet.register_stock_callback("preferences", STOCK_MENU_PREF, ("Preferences"), self.Preferences)
        self.applet.register_stock_callback("about", STOCK_MENU_ABOUT, ("About"), self.About)

        self.applet.set_events(self.applet.get_events() | GDK.BUTTON_PRESS_MASK | GDK.ENTER_NOTIFY_MASK | GDK.LEAVE_NOTIFY_MASK)
        self.applet.connect("button_press_event", self.ShowList)
        self.applet.connect("enter_notify_event", self.HighlightApplet)
        self.applet.connect("leave_notify_event", self.DarkenApplet)

        filename = self.appletpixmapmouseover.get_text()
        if (find(filename, "/") == -1):
            filename = "%s/%s" % (pixmappath, filename)
        self.pixbright, self.maskbright = create_pixmap_from_xpm(self.applet, None, filename)
        filename = self.appletpixmap.get_text()
        if (find(filename, "/") == -1):
            filename = "%s/%s" % (pixmappath, filename)
        self.pixdark, self.maskdark = create_pixmap_from_xpm(self.applet, None, filename)

        self.app_pix = GtkPixmap(self.pixdark, self.maskdark)
        self.app_pix.show()

        self.applet.add(self.app_pix)
        self.applet.show()

    # Create the main window (short & long versions)
    def CreateNonApplet(self, args):

        if (self.mainwindowstatus != MW_NONE):
            self.applet.destroy()
            if (self.mainwindowstatus == MW_SHORT):
                self.mainwindowstatus = MW_LONG
            else:
                self.mainwindowstatus = MW_SHORT
        else:
            self.mainwindowstatus = MW_SHORT

        self.applet = GtkWindow(WINDOW_POPUP)
        self.applet.set_uposition(atoi(self.nonappletpositionx.get_text()), atoi(self.nonappletpositiony.get_text()))
        handlebox = GtkHandleBox()
        hbox = GtkHBox()

        # Execute button
        execbutton = GtkButton()
        execbutton.add(self.CreatePixmap("execute.xpm"))
        execbutton.connect("clicked", self.ShowList)
        execbutton.show()

        # Create open / close buttons
        openclosebutton = GtkButton()
        if (self.mainwindowstatus == MW_SHORT):
            openclosebutton.add(self.CreatePixmap("open.xpm"))
        else:
            openclosebutton.add(self.CreatePixmap("close.xpm"))
        openclosebutton.connect("clicked", self.CreateNonApplet)
        openclosebutton.show()

        hbox.pack_start(execbutton)
        hbox.pack_start(openclosebutton)

        if (self.mainwindowstatus == MW_LONG):

            # Preferences button
            prefbutton = GtkButton()
            prefbutton.add(self.CreatePixmap("preferences.xpm"))
            prefbutton.connect("clicked", self.Preferences)
            prefbutton.show()

            # About button
            aboutbutton = GtkButton()
            aboutbutton.add(self.CreatePixmap("about.xpm"))
            aboutbutton.connect("clicked", self.About)
            aboutbutton.show()

            # Quit button
            quitbutton = GtkButton()
            quitbutton.add(self.CreatePixmap("quit.xpm"))
            quitbutton.connect("clicked", self.DestroyNonApplet)
            quitbutton.show()

            hbox.pack_start(prefbutton)
            hbox.pack_start(aboutbutton)
            hbox.pack_start(quitbutton)

        hbox.show()

        handlebox.add(hbox)
        handlebox.show()

        self.applet.add(handlebox)
        self.applet.show()

    # Hide the "Copy File" window
    def HideCopyFileWindow(self, *args):
        self.copyfile.hide()

    # Init the internal crypto routines
    def InitCrypto(self):
        self.cipher = twofish()
        self.cipher.hashkey(gethostname())

    # encrypt a string
    def encrypt(self, string):
        if string == "":
            return string
        else:
            newString = cfill(string, " ", (divmod(len(string), 16)[0]+1) * 16)
            newString = b2a_base64(self.cipher.encrypt_cbc(newString))
            return newString

    # decrypt a string
    def decrypt(self, string):
        if string == "":
            return string
        else:
            newString = self.cipher.decrypt_cbc(a2b_base64(string))
            newString = strip(newString)
            return newString

    # Handler for cleaning Zombie processes
    def CleanHandler(self):
        return TRUE

    # Close non-applet-window
    def DestroyNonApplet(self, args):
        timeout_remove(self.cleanhandler)
        self.applet.destroy()
        mainquit()

    # Ignore CopyFile-window destory / delete_event events
    def IgnoreDestroy(self, *args):
        return TRUE

    # Open a window for copying files to a remote host
    def OpenCopyFileWindow(self, filelist): 

        # Hide main window
        self.sshwin.hide()
        self.sshwindowmode = 2

        # Show "Copy File(s)" window
        self.directory.set_text(self.currenthost.defaultdirectory)
        self.copyfilelist.set_text(strip(filelist))
        self.hostinformation.set_text("%s [%s] - %s@%s" % (self.currenthost.name, self.currenthost.category, self.currenthost.username, self.currenthost.hostname))
        self.copyfile.show()

    # Destroy Directory Selection window with executing command
    def CopyFile(self, args):
        directory = self.directory.get_text()
        filelist = self.copyfilelist.get_text()
        self.currenthost.CopyFile(self.scpcommand.get_text(), self.resolve.active, directory, filelist, self.decrypt(self.currenthost.password))
        self.copyfile.hide()
        return TRUE

    # Destroy Directory Selection window
    def DestroyDirWin(self, args):
        self.dirEntry.save_history()
        self.opendirwin.destroy()

    # Event handler for listing mode
    def SelectUserHost(self, clist, row, column, event):
        name = clist.get_text(row, 3)
        category = clist.get_text(row, 4)
        clist.unselect_all()
        host = self.hosts[itemname(name, category)]

        self.sshwin.hide()
        self.sshwindowmode = 2

        host.Execute(self.sshcommand.get_text(), self.sshforwarding.get_text(), 
                     self.rshcommand.get_text(), self.telnetcommand.get_text(),
                     self.vncviewercommand.get_text(), self.resolve.active, self.decrypt(host.password))

    # Event handler for tree mode
    def SelectTree(self, ctree, row, column, event):
        node = ctree.node_nth(row)
        nodeinfo = ctree.get_node_info(node)
        name = nodeinfo[0]
        isleaf = nodeinfo[6]
        ctree.unselect_all()
        if (isleaf):
            category = ""
            for i in range(len(ctree.base_nodes())):
                basenode = ctree.base_nodes()[i]
                if (ctree.is_ancestor(basenode, node)):
                    category = ctree.get_node_info(basenode)[0]

            host = self.hosts[itemname(name, category)]

            self.sshwin.hide()
            self.sshwindowmode = 2

            host.Execute(self.sshcommand.get_text(), self.sshforwarding.get_text(), 
                         self.rshcommand.get_text(), self.telnetcommand.get_text(),
                         self.vncviewercommand.get_text(), self.resolve.active, self.decrypt(host.password))

    # Got some files to copy from the tree view
    def DragReceiveTree(self, ctree, context, x, y, data, info, time):
        y = y - (ctree.get_column_widget(0).size_request()[1] + 4)
        selection = ctree.get_selection_info(x, y)
        if (selection != None):
            row, col = selection
            filenames = split(data.data,"\n")
            filelist = ""
            for i in range(len(filenames)):
                if (filenames[i][:5] == "file:"):
                    filelist = "%s '%s'" % (filelist, strip(filenames[i][5:]))

            node = ctree.node_nth(row)
            nodeinfo = ctree.get_node_info(node)
            name = nodeinfo[0]
            isleaf = nodeinfo[6]
            ctree.unselect_all()
            if (isleaf):
                category = ""
                for i in range(len(ctree.base_nodes())):
                    basenode = ctree.base_nodes()[i]
                    if (ctree.is_ancestor(basenode, node)):
                        category = ctree.get_node_info(basenode)[0]

                self.currenthost = self.hosts[itemname(name, category)]
                self.OpenCopyFileWindow(filelist)

    # Got some files to copy from the list view
    def DragReceiveList(self, clist, context, x, y, data, info, time):
        y = y - (clist.get_column_widget(0).size_request()[1] + 4)
        selection = clist.get_selection_info(x, y)

        if (selection != None):
            row, col = selection
            filenames = split(data.data,"\n")
            filelist = ""
            for i in range(len(filenames)):
                if (filenames[i][:5] == "file:"):
                    filelist = "%s '%s'" % (filelist, strip(filenames[i][5:]))

            name = clist.get_text(row, 3)
            category = clist.get_text(row, 4)
            clist.unselect_all()

            self.currenthost = self.hosts[itemname(name, category)]
            self.OpenCopyFileWindow(filelist)

    # Show the list of available shells
    def ShowList(self, args, event = None):

        if (self.locked == TRUE):
            self.callingfunction = CF_SHOWLIST
            self.UnLock()
            return FALSE

        if (self.sshwindowmode == 1):
            self.sshwin.hide()
            self.sshwindowmode = 2
        elif (self.sshwindowmode == 2):
            self.sshwin.show()
            self.sshwindowmode = 1
        elif (self.sshwindowmode == 0):
            self.sshwindowmode = 1
            self.sshwin = GtkWindow(WINDOW_POPUP)
            self.sshwin.set_uposition(atoi(self.windowpositionx.get_text()), atoi(self.windowpositiony.get_text()))
            self.sshwin.set_usize(atoi(self.windowsizex.get_text()), atoi(self.windowsizey.get_text()))

            if (self.listview.active):
                self.ShowListView()
            elif (self.treeview.active):
                self.ShowTreeView()

            self.sshwin.show()

    # Create main window (list view)
    def ShowListView(self):
        swin = GtkScrolledWindow()
        swin.set_policy(POLICY_AUTOMATIC, POLICY_AUTOMATIC)

        self.connlist = GtkCList(5, [" Hostname ", " Username ", " Protocol ", " Name ", " Category "])
        self.connlist.set_column_visibility(4, FALSE)
        self.connlist.set_column_visibility(3, FALSE)
        self.connlist.column_titles_passive()
        self.connlist.set_column_width(0, atoi(self.windowsizex.get_text()) - 145)
        self.connlist.show()
        self.connlist.set_auto_sort(TRUE)
        self.connlist.connect("select_row", self.SelectUserHost)
        self.connlist.connect("drag_data_received", self.DragReceiveList)
        self.connlist.drag_dest_set(DEST_DEFAULT_ALL, targets, GDK.ACTION_COPY)

        for i in range(len(self.hosts.keys())):
            index = self.hosts.keys()[i]
            self.connlist.append([self.hosts[index].hostname, self.hosts[index].username, protocol2str(self.hosts[index].protocol), self.hosts[index].name, self.hosts[index].category])

        background = split(self.bgcolorone.get_text(), ",")
        rootcolor = split(self.rootcolor.get_text(), ",")
        for i in range(len(self.hosts.keys())):
            if (background == split(self.bgcolorone.get_text(), ",")):
                background = split(self.bgcolortwo.get_text(), ",")
            else:
                background = split(self.bgcolorone.get_text(), ",")
            self.connlist.set_background(i, GdkColor(atoi(background[0]), atoi(background[1]), atoi(background[2])))
            if (self.connlist.get_text(i, 1) == "root"):
                self.connlist.set_foreground(i, GdkColor(atoi(rootcolor[0]), atoi(rootcolor[1]), atoi(rootcolor[2])))

        swin.add(self.connlist)
        swin.show()

        self.sshwin.add(swin)        

    # Show main window (tree view)
    def ShowTreeView(self):
        swin = GtkScrolledWindow()
        swin.set_policy(POLICY_AUTOMATIC, POLICY_AUTOMATIC)

        categorylist = []
        for i in range(len(self.hosts.keys())):
            index = self.hosts.keys()[i]
            if (not self.hosts[index].category in categorylist):
                categorylist.append(self.hosts[index].category)
        categorylist.sort()

        pixopen, maskopen = create_pixmap_from_xpm(self.applet, None, "%s/folder-open.xpm" % pixmappath)
        pixclose, maskclose = create_pixmap_from_xpm(self.applet, None, "%s/folder.xpm" % pixmappath)
        pixssh, maskssh  = create_pixmap_from_xpm(self.applet, None, "%s/ssh.xpm" % pixmappath)
        pixrsh, maskrsh  = create_pixmap_from_xpm(self.applet, None, "%s/rsh.xpm" % pixmappath)
        pixtelnet, masktelnet  = create_pixmap_from_xpm(self.applet, None, "%s/telnet.xpm" % pixmappath)
        pixvnc, maskvnc = create_pixmap_from_xpm(self.applet, None, "%s/vnc.xpm" % pixmappath)

        self.conntree = GtkCTree(1, 0, [ " Connections "])
        self.conntree.column_titles_passive()
        self.conntree.connect("drag_data_received", self.DragReceiveTree)
        self.conntree.drag_dest_set(DEST_DEFAULT_ALL, targets, GDK.ACTION_COPY)
        self.conntree.connect("select_row", self.SelectTree)
        self.conntree.set_auto_sort(TRUE)
        for i in range(len(categorylist)):
            parent = self.conntree.insert_node(None, None, [ categorylist[i] ], is_leaf = FALSE, pixmap_closed = pixclose, pixmap_opened = pixopen, mask_closed = maskclose, mask_opened = maskopen)
            for j in range(len(self.hosts.keys())):
                index = self.hosts.keys()[j]
                if (self.hosts[index].category == categorylist[i]):
                    if (self.hosts[index].protocol == SSH):
                        self.conntree.insert_node(parent, None, [ self.hosts[index].name ], pixmap_closed = pixssh, mask_closed = maskssh)
                    elif (self.hosts[index].protocol == RSH):
                        self.conntree.insert_node(parent, None, [ self.hosts[index].name ], pixmap_closed = pixrsh, mask_closed = maskrsh)
                    elif (self.hosts[index].protocol == TELNET):
                        self.conntree.insert_node(parent, None, [ self.hosts[index].name ], pixmap_closed = pixtelnet, mask_closed = masktelnet)
                    elif (self.hosts[index].protocol == VNC):
                        self.conntree.insert_node(parent, None, [ self.hosts[index].name ], pixmap_closed = pixvnc, mask_closed = maskvnc)

        self.conntree.show()
        swin.add_with_viewport(self.conntree)
        swin.show()

        # Plus button
        plusbutton = GtkButton()
        plusbutton.add(self.CreatePixmap("plus.xpm"))
        plusbutton.connect("clicked", self.ExpandAll)
        plusbutton.show()

        # Minus button
        minusbutton = GtkButton()
        minusbutton.add(self.CreatePixmap("minus.xpm"))
        minusbutton.connect("clicked", self.CollapseAll)
        minusbutton.show()

        label = GtkLabel()
        label.show()

        hbox = GtkHBox()
        hbox.pack_start(label)
        hbox.pack_start(plusbutton, expand = FALSE)
        hbox.pack_start(minusbutton, expand = FALSE)
        hbox.show()

        vbox = GtkVBox()
        vbox.show()
        vbox.pack_start(swin)
        vbox.pack_start(hbox, expand = FALSE)

        self.sshwin.add(vbox)        
        
    # Expand all nodes in the CTree
    def ExpandAll(self, args):
        basenodes = self.conntree.base_nodes()
        for node in basenodes:
            self.conntree.expand_recursive(node)

    # Collapse all nodes in the CTree
    def CollapseAll(self, args):
        basenodes = self.conntree.base_nodes()
        for node in basenodes:
            self.conntree.collapse_recursive(node)

    # Define an about box
    def About(self, args):
        GnomeAbout(title, version, 'May be distributed under the terms of the GPL2', ['Martin Preishuber'], 'A simple shell for starting SSH.').show()

    # Create Help CList
    def CreateHelpCList(self):
        self.helplist.set_column_justification(0, JUSTIFY_CENTER)
        self.helplist.set_column_justification(2, JUSTIFY_CENTER)
        self.helplist.set_column_justification(3, JUSTIFY_CENTER)
        self.helplist.set_column_justification(4, JUSTIFY_CENTER)
        self.helplist.set_column_justification(5, JUSTIFY_CENTER)
        self.helplist.set_column_justification(6, JUSTIFY_CENTER)

        self.helplist.append([ "%a", "Application", "", "", "", "", "" ])
        self.helplist.set_pixmap(0, 2, self.YesXpm())
        self.helplist.set_pixmap(0, 3, self.YesXpm())
        self.helplist.set_pixmap(0, 4, self.YesXpm())

        self.helplist.append([ "%c", "Cipher", "", "", "", "", "" ])
        self.helplist.set_pixmap(1, 2, self.YesXpm())
        self.helplist.set_pixmap(1, 5, self.YesXpm())

        self.helplist.append([ "%C", "Compression", "", "", "", "", "" ])
        self.helplist.set_pixmap(2, 2, self.YesXpm())
        self.helplist.set_pixmap(2, 5, self.YesXpm())

        self.helplist.append([ "%D", "Target directory", "", "", "", "", "" ])
        self.helplist.set_pixmap(3, 5, self.YesXpm())

        self.helplist.append([ "%e", "Encodings", "", "", "", "", ""])
        self.helplist.set_pixmap(4, 6, self.YesXpm())

        self.helplist.append([ "%f", "Program to launch", "", "", "", "", "" ])
        self.helplist.set_pixmap(5, 2, self.YesXpm())

        self.helplist.append([ "%F", "File list", "", "", "", "", "" ])
        self.helplist.set_pixmap(6, 5, self.YesXpm())

        self.helplist.append([ "%h", "Hostname / Display", "", "", "", "", "" ])
        self.helplist.set_pixmap(7, 2, self.YesXpm())
        self.helplist.set_pixmap(7, 3, self.YesXpm())
        self.helplist.set_pixmap(7, 4, self.YesXpm())
        self.helplist.set_pixmap(7, 5, self.YesXpm())
        self.helplist.set_pixmap(7, 6, self.YesXpm())

        self.helplist.append([ "%L", "Local forwarding (port:host:hostport)", "", "", "", "", "" ])
        self.helplist.set_pixmap(8, 2, self.YesXpm())

        self.helplist.append([ "%p", "Port to use", "", "", "", "", "" ])
        self.helplist.set_pixmap(9, 2, self.YesXpm())
        self.helplist.set_pixmap(9, 4, self.YesXpm())
        self.helplist.set_pixmap(9, 5, self.YesXpm())

        self.helplist.append([ "%R", "Remote forwarding (port:host:hostport)", "", "", "", "", "" ])
        self.helplist.set_pixmap(10, 2, self.YesXpm())

        self.helplist.append([ "%s", "Fullscreen", "", "", "", "", ""])
        self.helplist.set_pixmap(11, 6, self.YesXpm())

        self.helplist.append([ "%S", "Shared", "", "", "", "", ""])
        self.helplist.set_pixmap(12, 6, self.YesXpm())

        self.helplist.append([ "%t", "Tunnel", "", "", "", "", ""])
        self.helplist.set_pixmap(13, 6, self.YesXpm())

        self.helplist.append([ "%u", "Username (e.g. root)", "", "", "", "", "" ])
        self.helplist.set_pixmap(14, 2, self.YesXpm())
        self.helplist.set_pixmap(14, 3, self.YesXpm())
        self.helplist.set_pixmap(14, 4, self.YesXpm())
        self.helplist.set_pixmap(14, 5, self.YesXpm())

        self.helplist.append([ "%v", "View only", "", "", "", "", "" ])
        self.helplist.set_pixmap(15, 6, self.YesXpm())
        
        self.helplist.append([ "%x", "X11 forwarding", "", "", "", "", "" ])
        self.helplist.set_pixmap(16, 2, self.YesXpm())

        self.helplist.append([ "%z", "ZLib compression-level", "", "", "", "", "" ])
        self.helplist.set_pixmap(17, 6, self.YesXpm())

    # Define the preferences window
    def Preferences(self, args):

        if (self.locked == TRUE):
            self.callingfunction = CF_PREFERENCES
            self.UnLock()
            return FALSE

        self.ClearPreferencesHostList()
        self.ShowPreferencesHostList()
        self.prefwin.show()
        self.prefasksavequestion = TRUE

    # Add a host to the hostlist
    def AddHost(self, args):
        self.currenthost = host(self.defaultcipher.get_menu().children().index(self.defaultcipher.get_menu().get_active()))
        self.SetHostPreferences()
        self.hostwin.show()

    # Set preferences in the host preferences tab for a given host
    def SetHostPreferences(self):

        # Items in the "General" tab
        self.connectionname.set_text(self.currenthost.name)
        self.category.set_text(self.currenthost.category)
        self.protocol.get_menu().set_active(self.currenthost.protocol)
        self.protocol.get_menu().get_active().select()
        self.hostname.set_text(self.currenthost.hostname)
        self.internalpassword.set_active(self.currenthost.internalpassword)
        self.password.set_text(self.decrypt(self.currenthost.password))
        self.username.set_text(self.currenthost.username)
        self.startapplication.set_text(self.currenthost.application)

        # Items in the "SSH" tab
        self.launch.set_text(self.currenthost.launch)
        self.localforwarding.set_text(self.currenthost.localforwarding)
        self.remoteforwarding.set_text(self.currenthost.remoteforwarding)
        self.sshport.set_text(str(self.currenthost.sshport))
        self.sshcipher.get_menu().set_active(self.currenthost.cipher)
        self.sshcipher.get_menu().get_active().select()
        self.x11forwarding.set_active(self.currenthost.x11forwarding)
        self.compression.set_active(self.currenthost.compression)
        self.scpdirectory.set_text(self.currenthost.defaultdirectory)

        # Items in the "Telnet" tab
        self.telnetport.set_text(str(self.currenthost.telnetport))

        # Items in the "VNC" tab
        self.encodings.set_text(self.currenthost.encodings)
        self.zlib.set_text(self.currenthost.zlib)
        self.viewonly.set_active(self.currenthost.viewonly)
        self.fullscreen.set_active(self.currenthost.fullscreen)
        self.shared.set_active(self.currenthost.shared)
        self.tunnel.set_active(self.currenthost.tunnel)

    # Remove a host from the hostlist
    def RemoveHost(self, args):
        if (len(self.prefhostlist.selection) != 1):
            GnomeErrorDialog("You have to select a host first !")
            return
        else:
            name = self.prefhostlist.get_text(self.prefhostlist.selection[0], 0)
            category = self.prefhostlist.get_text(self.prefhostlist.selection[0], 1)

        host = self.hosts[itemname(name, category)]
        del self.hosts[itemname(name, category)]
        self.ClearPreferencesHostList()
        self.ShowPreferencesHostList()

    # Update a host in the list
    def UpdateHost(self, args):
        if (len(self.prefhostlist.selection) != 1):
            GnomeErrorDialog("You have to select a host first !")
            return
        else:
            name = self.prefhostlist.get_text(self.prefhostlist.selection[0], 0)
            category = self.prefhostlist.get_text(self.prefhostlist.selection[0], 1)

        host = self.hosts[itemname(name, category)]
        del self.hosts[itemname(name, category)]
        self.currenthost = host
        self.SetHostPreferences()
        self.hostwin.show()

    # Clear the host-list in the preferences window
    def ClearPreferencesHostList(self):
        self.prefhostlist.clear()

    # Show list of available hosts in the preferences window
    def ShowPreferencesHostList(self):
        for i in range(len(self.hosts.keys())):
            key = self.hosts.keys()[i]
            self.prefhostlist.append([ self.hosts[key].name, self.hosts[key].category, self.hosts[key].hostname ])

    # Answer on the "Save Preferences" question
    def PreferencesSaveQuestion(self, answer, *args):
        self.prefwin.hide()
        if (answer == YES):
            self.SavePreferences()
        else:
            self.LoadPreferences()

    # Close the preferences window
    def PreferencesWindowHide(self, *args):
        if self.prefasksavequestion == TRUE:
            self.prefasksavequestion = FALSE
            GnomeQuestionDialog("Save preferences ?", self.PreferencesSaveQuestion)
        return TRUE

    # Close the host preferences window
    def HostPreferencesWindowHide(self, *args):

        if (self.connectionname.get_text() == "") or (self.category.get_text() == ""):
            GnomeErrorDialog("You have to specify at least a connection name and category !")
            return TRUE
        
        # Set general settings for the current host
        self.currenthost.name = self.connectionname.get_text()
        self.currenthost.category = self.category.get_text()
        self.currenthost.protocol = self.protocol.get_menu().children().index(self.protocol.get_menu().get_active())
        self.currenthost.hostname = self.hostname.get_text()
        self.currenthost.internalpassword = self.internalpassword.active
        self.currenthost.password = self.encrypt(self.password.get_text())
        self.currenthost.username = self.username.get_text()
        self.currenthost.application = self.startapplication.get_text()

        # Set "SSH" settings for the current host
        self.currenthost.launch = self.launch.get_text()
        self.currenthost.localforwarding = self.localforwarding.get_text()
        self.currenthost.remoteforwarding = self.remoteforwarding.get_text()
        self.currenthost.sshport = atoi(self.sshport.get_text())
        self.currenthost.cipher = self.sshcipher.get_menu().children().index(self.sshcipher.get_menu().get_active())
        self.currenthost.x11forwarding = self.x11forwarding.active
        self.currenthost.compression = self.compression.active
        self.currenthost.defaultdirectory = self.scpdirectory.get_text()

        # Set "Telnet" settings for the current host
        self.currenthost.telnetport = atoi(self.telnetport.get_text())

        # Set "VNC" settings for the current host
        self.currenthost.encodings = self.encodings.get_text()
        self.currenthost.zlib = self.zlib.get_text()
        self.currenthost.viewonly = self.viewonly.active
        self.currenthost.fullscreen = self.fullscreen.active
        self.currenthost.shared = self.shared.active
        self.currenthost.tunnel = self.tunnel.active

        self.hosts[itemname(self.currenthost.name, self.currenthost.category)] = self.currenthost

        self.hostwin.hide()

        self.ClearPreferencesHostList()
        self.ShowPreferencesHostList()

        return TRUE

    # Load options
    def LoadPreferences(self):
        parser = ConfigParser()                             # The Configuration File Parser
        filename = "%s/%s" % (personaldirectory, ressourcefile)
        
        try:
            parser.read(filename)
            
            for i in range(len(parser.options("esshsh"))):
                option = parser.options("esshsh")[i]
                value = parser.get("esshsh", option)

                # Set settings in the "Command" tab
                if (option == "command"): self.sshcommand.set_text(value)
                elif (option == "forward"): self.sshforwarding.set_text(value)
                elif (option == "rsh"): self.rshcommand.set_text(value)
                elif (option == "scp"): self.scpcommand.set_text(value)
                elif (option == "telnet"): self.telnetcommand.set_text(value)
                elif (option == "vnc"): self.vncviewercommand.set_text(value)
                    
                # Set settings in the "Window Size & Position" tab
                elif (option == "posx"): self.windowpositionx.set_text(value)
                elif (option == "posy"): self.windowpositiony.set_text(value)
                elif (option == "sizex"): self.windowsizex.set_text(value)
                elif (option == "sizey"): self.windowsizey.set_text(value)
                elif (option == "appletsizex"): self.appletsizex.set_text(value)
                elif (option == "appletsizey"): self.appletsizey.set_text(value)
                elif (option == "appletpixmap"): self.appletpixmap.set_text(value)
                elif (option == "appletpixmapmouseover"): self.appletpixmapmouseover.set_text(value)
                elif (option == "bgone"): self.bgcolorone.set_text(value)
                elif (option == "bgtwo"): self.bgcolortwo.set_text(value)
                elif (option == "rootcolor"): self.rootcolor.set_text(value)
                elif (option == "viewmode"): 
                    if (str2viewmode(value) == LISTVIEW):
                        self.listview.set_active(1)
                    elif (str2viewmode(value) == TREEVIEW):
                        self.treeview.set_active(1)
                elif (option == "mainposx"): self.nonappletpositionx.set_text(value)
                elif (option == "mainposy"): self.nonappletpositiony.set_text(value)

                # set settings in the "Miscellaneous" tab
                elif (option == "resolve"): self.resolve.set_active(str2bool(value))
                elif (option == "defaultcipher"): 
                    self.defaultcipher.get_menu().set_active(str2cipher(value))
                    self.defaultcipher.get_menu().get_active().select()

            # Load hosts
            for i in range(len(parser.options("hosts"))):
                option = parser.options("hosts")[i]
                value = parser.get("hosts", option)

                if (option[:4] == "host"):
                    newhost = host()
                    newhost.Load(value)
                    self.hosts[itemname(newhost.name, newhost.category)] = newhost

        except:
            self.log4py.warn("Error while parsing configuration file - setting default values.")

            # Set default values in the "Command" tab
            self.sshcommand.set_text(defaultcommand)
            self.sshforwarding.set_text(defaultforward)
            self.rshcommand.set_text(defaultrsh)
            self.scpcommand.set_text(defaultscp)
            self.telnetcommand.set_text(defaulttelnet)
            self.vncviewercommand.set_text(defaultvncviewer)

            # Set default values in the "Window Size & Position" tab
            self.windowpositionx.set_text(str(defaultwindowx))
            self.windowpositiony.set_text(str(defaultwindowy))
            self.windowsizex.set_text(str(defaultwidthx))
            self.windowsizey.set_text(str(defaultwidthy))
            self.appletsizex.set_text(str(defaultappletsizex))
            self.appletsizey.set_text(str(defaultappletsizey))
            self.appletpixmap.set_text(str(defaultappletpixmap))
            self.appletpixmapmouseover.set_text(str(defaultappletpixmapmouseover))
            self.bgcolorone.set_text(join(defaultbgone, ","))
            self.bgcolortwo.set_text(join(defaultbgtwo, ","))
            self.rootcolor.set_text(join(defaultcolorroot, ","))
            self.treeview.set_active(TRUE)
            self.nonappletpositionx.set_text(str(defaultnonappletposx))
            self.nonappletpositiony.set_text(str(defaultnonappletposy))

            # Set default values in the "Miscellaneous" tab
            self.resolve.set_active(TRUE)
            self.defaultcipher.get_menu().set_active(0)
            self.defaultcipher.get_menu().get_active().select()

            self.sshwindowmode = defaultsshwinmode

    # Save options
    def SavePreferences(self):

        filename = "%s/%s" % (personaldirectory, ressourcefile)
        file = open(filename, "w")
        file.write("[esshsh]\n")
        
        # Save values in the "Command" tab
        file.write("Command=%s\n" % self.sshcommand.get_text())
        file.write("Forward=%s\n" % self.sshforwarding.get_text())
        file.write("Rsh=%s\n" % self.rshcommand.get_text())
        file.write("Telnet=%s\n" % self.telnetcommand.get_text())
        file.write("Scp=%s\n" % self.scpcommand.get_text())
        file.write("VNC=%s\n" % self.vncviewercommand.get_text())
        
        # Save values in the "Window Size & Position" tab
        file.write("PosX=%s\n" % self.windowpositionx.get_text())
        file.write("PosY=%s\n" % self.windowpositiony.get_text())
        file.write("SizeX=%s\n" % self.windowsizex.get_text())
        file.write("SizeY=%s\n" % self.windowsizey.get_text())
        file.write("AppletSizeX=%s\n" % self.appletsizex.get_text())
        file.write("AppletSizeY=%s\n" % self.appletsizey.get_text())
        file.write("AppletPixmap=%s\n" % self.appletpixmap.get_text())
        file.write("AppletPixmapMouseOver=%s\n" % self.appletpixmapmouseover.get_text())
        file.write("BgOne=%s\n" % self.bgcolorone.get_text())
        file.write("BgTwo=%s\n" % self.bgcolortwo.get_text())
        file.write("RootColor=%s\n" % self.rootcolor.get_text())
        if (self.treeview.active): file.write("ViewMode=%s\n" % viewmode2str(TREEVIEW))
        elif (self.listview.active): file.write("ViewMode=%s\n" % viewmode2str(LISTVIEW))
        file.write("MainPosX=%s\n" % self.nonappletpositionx.get_text())
        file.write("MainPosY=%s\n" % self.nonappletpositiony.get_text())

        # Save values from the "Miscellaneous" tab
        file.write("Resolve=%s\n" % bool2str(self.resolve.active))
        file.write("DefaultCipher=%s\n" % cipher2str(self.defaultcipher.get_menu().children().index(self.defaultcipher.get_menu().get_active())))

        file.write("\n[hosts]\n")
        for i in range(len(self.hosts.keys())):
            self.hosts[self.hosts.keys()[i]].Save()
            file.write("Host%d=%s\n" % (i, self.hosts[self.hosts.keys()[i]].filename))
        file.close()

        if (self.sshwindowmode == 2) or (self.sshwindowmode == 1):
            self.sshwin.destroy()
            self.sshwindowmode = 0

    # Returns a pixmap from a file
    def CreatePixmap(self, filename):
        pix, mask = create_pixmap_from_xpm(self.applet, None, "%s/%s" % (pixmappath, filename))
        pixmap = GtkPixmap(pix, mask)
        pixmap.show()
        return pixmap

    # Creates a pixmap indicating "yes"
    def YesXpm(self):
        return self.CreatePixmap("yes.xpm")

    # Main program
    def Main(self):
        self.cleanhandler = timeout_add(1000, self.CleanHandler)
        mainloop()

# -----------------------------------------------------------------------------
# That's all folks ;-)
